
/*
    pbrt source code is Copyright(c) 1998-2016
                        Matt Pharr, Greg Humphreys, and Wenzel Jakob.

    This file is part of pbrt.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are
    met:

    - Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    - Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
    IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
    TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

 */

// lights/diffuse.cpp*
#include "lights/textured.h"
#include "paramset.h"
#include "sampling.h"
#include "shapes/triangle.h"
#include "stats.h"
#include "imageio.h"

namespace pbrt {

// TexturedAreaLight Method Definitions
TexturedAreaLight::TexturedAreaLight(const Transform &LightToWorld,
                                   const MediumInterface &mediumInterface,
								   const Spectrum &Le, const std::string &texmap, int nSamples,
                                   const std::shared_ptr<Shape> &shape,
                                   bool twoSided)
    : AreaLight(LightToWorld, mediumInterface, nSamples),
      shape(shape),
      twoSided(twoSided),
      area(shape->Area()) {
    // Warn if light has transformation with non-uniform scale, though not
    // for Triangles, since this doesn't matter for them.
    if (WorldToLight.HasScale() &&
        dynamic_cast<const Triangle *>(shape.get()) == nullptr)
        Warning(
            "Scaling detected in world to light transformation! "
            "The system has numerous assumptions, implicit and explicit, "
            "that this transform will have no scale factors in it. "
            "Proceed at your own risk; your image may have errors.");

	// Read texel data from _texmap_ and initialize _Lmap_
	Point2i resolution;
	std::unique_ptr<RGBSpectrum[]> texels(nullptr);
	if (texmap != "") {
		texels = ReadImage(texmap, &resolution);
	}
	if (texels) {
		avgLemit = Spectrum(0.f);
		for (int i = 0; i < resolution.x * resolution.y; ++i) {
			texels[i] *= Le.ToRGBSpectrum();
			avgLemit += Spectrum(texels[i], SpectrumType::Illuminant);
		}
		avgLemit /= resolution.x * resolution.y;
	} else {
		resolution.x = resolution.y = 1;
		texels = std::unique_ptr<RGBSpectrum[]>(new RGBSpectrum[1]);
		texels[0] = Le.ToRGBSpectrum();
		avgLemit = Le;
	}
	Lmap.reset(new MIPMap<RGBSpectrum>(resolution, texels.get(), false, 8.f, ImageWrap::Clamp));

	// Initialize sampling PDFs for infinite area light

	// Compute scalar-valued image _img_ from environment map
	int width = 2 * Lmap->Width(), height = 2 * Lmap->Height();
	std::unique_ptr<Float[]> img(new Float[width * height]);
	float fwidth = 0.5f / std::min(width, height);
	ParallelFor(
		[&](int64_t v) {
		Float vp = (v + .5f) / (Float)height;
		for (int u = 0; u < width; ++u) {
			Float up = (u + .5f) / (Float)width;
			img[u + v * width] = Lmap->Lookup(Point2f(up, vp), fwidth).y();
		}
	},
		height, 32);

	// Compute sampling distributions for rows and columns of image
	distribution.reset(new Distribution2D(img.get(), width, height));
}

Spectrum TexturedAreaLight::L(const Interaction &isectLight, const Vector3f &dirFromLight) const {
	if (twoSided || Dot(isectLight.n, dirFromLight) > 0.f) {
		const Point2f mapCoord = shape->PointToTextureCoord(isectLight.p);
		return Spectrum(Lmap->Lookup(mapCoord), SpectrumType::Illuminant);
	} else {
		return Spectrum(0.f);
	}
}

Spectrum TexturedAreaLight::Power() const {
	return (twoSided ? 2 : 1) * avgLemit * area * Pi;
}

Spectrum TexturedAreaLight::Sample_Li(const Interaction &ref, const Point2f &u,
                                     Vector3f *wi, Float *pdf,
                                     VisibilityTester *vis) const {
    ProfilePhase _(Prof::LightSample);

	*pdf = 0.f;

	// Sample position in the texture
	/*Float mapPdf = 1.f;
	Point2f mapSample = u;*/
	Float mapPdf;
	Point2f mapSample = distribution->SampleContinuous(u, &mapPdf);
	if (mapPdf == 0.f) {
		return Spectrum(0.f);
	}

	// Convert the position in the texture to a position on the shape
	Float shapePdf;
	Interaction shapeSample = shape->Sample(ref, mapSample, &shapePdf);
	shapeSample.mediumInterface = mediumInterface;
	if (shapePdf == 0.f || (shapeSample.p - ref.p).LengthSquared() == 0) {
		return Spectrum(0.f);
	}

	*wi = Normalize(shapeSample.p - ref.p);
	if (AbsDot(shapeSample.n, -*wi) == 0.f) {
		return Spectrum(0.f);
	}

	*vis = VisibilityTester(ref, shapeSample);
	*pdf = mapPdf / area * DistanceSquared(ref.p, shapeSample.p) / AbsDot(shapeSample.n, -*wi);

	if (twoSided || Dot(shapeSample.n, -*wi) > 0.f) {
		return Spectrum(Lmap->Lookup(mapSample), SpectrumType::Illuminant);
	}
	else {
		return Spectrum(0.f);
	}
}

Float TexturedAreaLight::Pdf_Li(const Interaction &ref,
                               const Vector3f &wi) const {
    ProfilePhase _(Prof::LightPdf);
	//return shape->Pdf(ref, wi);
	Ray ray = ref.SpawnRay(wi);
	Float tHit;
	SurfaceInteraction isectLight;
	if (!shape->Intersect(ray, &tHit, &isectLight, false) || AbsDot(isectLight.n, -wi) == 0.f) return 0.f;

	const Point2f mapCoord = shape->PointToTextureCoord(isectLight.p);
	const Float mapPdf = distribution->Pdf(mapCoord) / area;

	return mapPdf * DistanceSquared(ref.p, isectLight.p) / AbsDot(isectLight.n, -wi);
}

Spectrum TexturedAreaLight::Sample_Le(const Point2f &u1, const Point2f &u2,
                                     Float time, Ray *ray, Normal3f *nLight,
                                     Float *pdfPos, Float *pdfDir) const {
	LOG(ERROR) << StringPrintf("Unimplemented TexturedAreaLight::Sample_Le method called.");
	return Spectrum(0.f);
}

void TexturedAreaLight::Pdf_Le(const Ray &ray, const Normal3f &n, Float *pdfPos,
                              Float *pdfDir) const {
	LOG(ERROR) << StringPrintf("Unimplemented TexturedAreaLight::Pdf_Le method called.");
}

Bounds3f TexturedAreaLight::Bounds() const {
    return shape->WorldBound();
}

std::shared_ptr<AreaLight> CreateTexturedAreaLight(
    const Transform &light2world, const Medium *medium,
    const ParamSet &paramSet, const std::shared_ptr<Shape> &shape) {
    Spectrum L = paramSet.FindOneSpectrum("L", Spectrum(1.0));
    Spectrum sc = paramSet.FindOneSpectrum("scale", Spectrum(1.0));
	std::string texmap = paramSet.FindOneFilename("mapname", "");
    int nSamples = paramSet.FindOneInt("samples",
                                       paramSet.FindOneInt("nsamples", 1));
    bool twoSided = paramSet.FindOneBool("twosided", false);
    if (PbrtOptions.quickRender) nSamples = std::max(1, nSamples / 4);
	return std::make_shared<TexturedAreaLight>(light2world, medium, L * sc, texmap,
                                              nSamples, shape, twoSided);
}

}  // namespace pbrt
